#define KWARE_C
#include "kware.h"

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <sys/io.h>
#include <sys/types.h>

#if 0

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/sched.h>

#else

struct file_operations;

#endif


#include <linux/module.h>

extern int     init_module(void);
extern void cleanup_module(void);

struct module __this_module;

int printk(const char * fmt, ...) __attribute__((format(printf,1,2)));
int printk(const char * fmt, ...)
{
va_list ap;
int i;

	va_start(ap,fmt);
	i=vprintf(fmt,ap);
	va_end(ap);
	return(i);
}

struct file_operations *module_fops;
struct file             module_file;
struct inode            module_inode;

int check_region(unsigned long from, unsigned long extent)
{ printf("check_region(from=0x%lX, extent=0x%lX) = 0\n",from,extent); return(0); }

void request_region(unsigned long from, unsigned long extent,const char *name)
{ printf("request_region(from=0x%lX, extent=0x%lX, name=\"%s\")\n",from,extent,name); }

void release_region(unsigned long from, unsigned long extent)
{ printf("release_region(from=0x%lX, extent=0x%lX)\n",from,extent); }

asmlinkage void schedule(void)
{ printf("schedule()\n"); }

int register_chrdev(unsigned int major, const char *name, struct file_operations *fops)
{ printf("register_chrdev(major=%d, name=\"%s\", fops=%p) = 0\n",major,name,(module_fops=fops)); return(0); }

int unregister_chrdev(unsigned int major, const char * name)
{ printf("unregister_chrdev(major=%d, name=\"%s\") = 0\n",major,name); return(0); }

int kware_open(const char *pathname, int flags)
{
int r;

	printf("   kware_open(pathname=\"%s\",flags=0x%x)...\n",pathname,flags);
	printf("   module_fops->open(inode,file)\n");
	r=module_fops->open(&module_inode,&module_file);
	printf("...module_fops->open(inode,file) = %d\n",r);
	if (r<0) { errno=-r; r=-1; }
	else r=10;
	printf("...kware_open(pathname=\"%s\",flags=0x%x) = %d\n",pathname,flags,r);
	return(r);
}

#define READWRITE(RW,RWS) \
 \
ssize_t kware_##RW(int fd, void *buf, size_t count) \
{ \
int r; \
loff_t offs=0; \
 \
	printf("   kware_" RWS "(fd=%d,buf=%p,count=%u)...\n",fd,buf,count); \
	printf("   module_fops->" RWS "(file,buf=%p,count=%u,*ppos=0)...",buf,count); \
	prepuser(); \
	r=module_fops->RW(&module_file,buf,count,&offs); \
	printf("...module_fops->" RWS "(file,buf=%p,count=%u) = (*ppos=%Ld) %d",buf,count,offs,r); \
	printf("...kware_" RWS "(fd=%d,buf=%p,count=%u) = %d\n",fd,buf,count,r); \
	if (r<0) { errno=-r; r=-1; } \
	return(r); \
}

READWRITE(read,"read")
READWRITE(write,"write")

int kware_ioctl(int fd,int cmd,unsigned long arg)
{
int r;

	printf("   kware_ioctl(fd=%d,cmd=%d,arg=%lu)...\n",fd,cmd,arg);
	printf("   module_fops->ioctl(inode,file,cmd=%d,arg=%lu)\n",cmd,arg);
	prepuser();
	r=module_fops->ioctl(&module_inode,&module_file,cmd,arg);
	printf("...module_fops->ioctl(inode,file,cmd=%d,arg=%lu) = %d\n",cmd,arg,r);
	printf("...kware_ioctl(fd=%d,cmd=%d,arg=%lu) = %d\n",fd,cmd,arg,r);
	if (r<0) { errno=-r; r=-1; }
	return(r);
}

extern void user_main(int argc,char **argv);

int main(int argc,char **argv)
{
int r;

	puts(  "   iopl(3)...");
	r=iopl(3);
	printf("...iopl(3) = %d\n",r);
	if (r) return(-r);
	puts(  "   init_module()...");
	r=init_module();
	printf("...init_module()   = %d\n",r);
	if (r) return(-r);
	puts(  "   user_main()...");
	user_main(argc,argv);
	puts(  "...user_main()");
	puts(  "   cleanup_module()...");
	cleanup_module();
	puts(  "...cleanup_module()"   );
	return(0);
}
